@@ -42,6 +42,7 @@
 #include "px_error.h"
 #include "px_misc.h"
 #include "px_encode.h"
+#include "px_crypt.h"
 
 #define max(a,b) ((a)>(b) ? (a) : (b))
 
@@ -175,6 +176,7 @@ PX_new3(void  (*errorhandler)(pxdoc_t *p, int type, const char *msg, void *data)
 	pxdoc->inputencoding = NULL;
 	pxdoc->px_data = NULL;
 	pxdoc->px_datalen = 0;
+	pxdoc->curblocknr = 0;
 
 	return pxdoc;
 }
@@ -222,10 +224,10 @@ PX_get_opaque(pxdoc_t *pxdoc) {
  */
 PXLIB_API int PXLIB_CALL
 PX_set_io_stream(pxdoc_t *pxdoc,
-                 char * (*readproc)(pxdoc_t *p, void *stream, size_t numbytes, void *buffer),
-                 size_t (*writeproc)(pxdoc_t *p, void *stream, const char *data, size_t len),
-                 size_t (*seekproc)(pxdoc_t *p, void *stream, size_t offset, int whence),
-                 size_t (*tellproc)(pxdoc_t *p, void *stream)
+                 size_t (*readproc)(pxdoc_t *p, pxstream_t *stream, size_t numbytes, void *buffer),
+                 size_t (*writeproc)(pxdoc_t *p, pxstream_t *stream, size_t len, void *data),
+                 int (*seekproc)(pxdoc_t *p, pxstream_t *stream, long offset, int whence),
+                 long (*tellproc)(pxdoc_t *p, pxstream_t *stream)
 				 ) {
 	pxstream_t *pxs;
 
@@ -234,8 +236,8 @@ PX_set_io_stream(pxdoc_t *pxdoc,
 		return -1;
 	}
 
-	if(NULL == (pxs = pxdoc->malloc(pxdoc, sizeof(pxstream_t), _("Allocate memory for io stream.")))) {
-		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for io stream."));
+	if(NULL == (pxs = px_stream_new(pxdoc))) {
+		px_error(pxdoc, PX_MemoryError, _("Could not create new io stream."));
 		return -1;
 	}
 	
@@ -359,27 +361,22 @@ PX_open_gsf(pxdoc_t *pxdoc, GsfInput *gsf) {
 		return -1;
 	}
 
-	if(NULL == (pxs = pxdoc->malloc(pxdoc, sizeof(pxstream_t), _("Allocate memory for io stream.")))) {
-		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for io stream."));
+	if(NULL == (pxs = px_stream_new_gsf(pxdoc, pxfFileRead, px_false, gsf))) {
+		px_error(pxdoc, PX_MemoryError, _("Could not create new gsf io stream."));
 		return -1;
 	}
-	pxs->type = pxfIOGsf;
-	pxs->mode = pxfFileRead;
-	pxs->close = px_false;
-	pxs->s.gsfin = gsf;
-	
-	pxdoc->read = px_gsfread;
-	pxdoc->seek = px_gsfseek;
-	pxdoc->tell = px_gsftell;
-	pxdoc->write = px_gsfwrite;
+	pxdoc->px_stream = pxs;
+
+	pxdoc->read = px_read;
+	pxdoc->seek = px_seek;
+	pxdoc->tell = px_tell;
+	pxdoc->write = px_write;
 
 	if((pxdoc->px_head = get_px_head(pxdoc, pxs)) == NULL) {
 		px_error(pxdoc, PX_RuntimeError, _("Unable to get header."));
 		return -1;
 	}
 
-	pxdoc->px_stream = pxs;
-
 	/* Build primary index. This index misses all index blocks with a level
 	 * greater than 1. Since they are not used currently this is of no harm.
 	 */
@@ -408,27 +405,23 @@ PX_open_fp(pxdoc_t *pxdoc, FILE *fp) {
 		return -1;
 	}
 
-	if(NULL == (pxs = pxdoc->malloc(pxdoc, sizeof(pxstream_t), _("Allocate memory for io stream.")))) {
-		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for io stream."));
+	if(NULL == (pxs = px_stream_new_file(pxdoc, pxfFileRead, px_false, fp))) {
+		px_error(pxdoc, PX_MemoryError, _("Could not create new file io stream."));
 		return -1;
 	}
-	pxs->type = pxfIOFile;
-	pxs->mode = pxfFileRead;
-	pxs->close = px_false;
-	pxs->s.fp = fp;
-	
-	pxdoc->read = px_fread;
-	pxdoc->seek = px_fseek;
-	pxdoc->tell = px_ftell;
-	pxdoc->write = px_fwrite;
+
+	pxdoc->px_stream = pxs;
+
+	pxdoc->read = px_read;
+	pxdoc->seek = px_seek;
+	pxdoc->tell = px_tell;
+	pxdoc->write = px_write;
 
 	if((pxdoc->px_head = get_px_head(pxdoc, pxs)) == NULL) {
 		px_error(pxdoc, PX_RuntimeError, _("Unable to get header."));
 		return -1;
 	}
 
-	pxdoc->px_stream = pxs;
-
 	/* Build primary index. This index misses all index blocks with a level
 	 * greater than 1. Since they are not used currently this is of no harm.
 	 */
@@ -510,6 +503,7 @@ PX_create_fp(pxdoc_t *pxdoc, pxfield_t *fields, int numfields, FILE *fp, int typ
 	if(type == pxfFileTypPrimIndex)
 		pxh->px_autoinc = 1;
 	pxh->px_sortorder = 0x62;
+	pxh->px_encryption = 0;
 
 	/* Calculate record size and get an idea on how big the header might
 	 * get due to the fieldnames, which is the major none fixed size. */
@@ -561,19 +555,17 @@ PX_create_fp(pxdoc_t *pxdoc, pxfield_t *fields, int numfields, FILE *fp, int typ
 	pxh->px_headersize = ((approxheadersize / 0x800) + 1) * 0x0800;
 //	fprintf(stderr, "Set headersize to 0x%X (%d)\n", pxh->px_headersize, approxheadersize);
 
-	if(NULL == (pxs = pxdoc->malloc(pxdoc, sizeof(pxstream_t), _("Allocate memory for io stream.")))) {
-		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for io stream."));
+	if(NULL == (pxs = px_stream_new_file(pxdoc, pxfFileWrite, px_false, fp))) {
+		px_error(pxdoc, PX_MemoryError, _("Could not create new file io stream."));
 		return -1;
 	}
-	pxs->type = pxfIOFile;
-	pxs->mode = pxfFileWrite;
-	pxs->close = px_false;
-	pxs->s.fp = fp;
-	
-	pxdoc->read = px_fread;
-	pxdoc->seek = px_fseek;
-	pxdoc->tell = px_ftell;
-	pxdoc->write = px_fwrite;
+
+	pxdoc->px_stream = pxs;
+
+	pxdoc->read = px_read;
+	pxdoc->seek = px_seek;
+	pxdoc->tell = px_tell;
+	pxdoc->write = px_write;
 
 	if(put_px_head(pxdoc, pxh, pxs) < 0) {
 		px_error(pxdoc, PX_RuntimeError, _("Unable to put header."));
@@ -581,7 +573,6 @@ PX_create_fp(pxdoc_t *pxdoc, pxfield_t *fields, int numfields, FILE *fp, int typ
 	}
 
 	pxdoc->px_head = pxh;
-	pxdoc->px_stream = pxs;
 	return 0;
 }
 /* }}} */
@@ -722,6 +713,9 @@ PX_get_value(pxdoc_t *pxdoc, const char *name, float *value) {
 	} else if(strcmp(name, "sortorder") == 0) {
 		*value = (float) pxdoc->px_head->px_sortorder;
 		return(0);
+	} else if(strcmp(name, "encryption") == 0) {
+		*value = (float) pxdoc->px_head->px_encryption;
+		return(0);
 	}
 	px_error(pxdoc, PX_Warning, _("No such value name."));
 	return(-2);
@@ -756,6 +750,16 @@ PX_set_parameter(pxdoc_t *pxdoc, const char *name, const char *value) {
 			px_error(pxdoc, PX_Warning, _("File is not writable. Setting '%s' has no effect."), name);
 			return -1;
 		}
+	} else if(strcmp(name, "password") == 0) {
+		pxdoc->px_head->px_encryption = px_passwd_checksum(value);
+		if(pxdoc->px_stream->mode & pxfFileWrite) {
+			if(put_px_head(pxdoc, pxdoc->px_head, pxdoc->px_stream) < 0) {
+				return -1;
+			}
+		} else {
+			px_error(pxdoc, PX_Warning, _("File is not writable. Setting '%s' has no effect."), name);
+			return -1;
+		}
 	} else if(strcmp(name, "targetencoding") == 0) {
 		int codepage;
 #if PX_USE_RECODE || PX_USE_ICONV
@@ -1090,7 +1094,7 @@ PX_write_primary_index(pxdoc_t *pxdoc, pxdoc_t *pxindex) {
 	 * we will need level 2 entries.
 	 * There is currently no support for level 3 entries.
 	 */
-	if(pih->px_maxtablesize*0x400-sizeof(TDataBlock) < indexdatalen*pih->px_recordsize) {
+	if(pih->px_maxtablesize*0x400-(int)sizeof(TDataBlock) < indexdatalen*pih->px_recordsize) {
 		pih->px_numindexlevels = 2;
 		recsperblock = (pih->px_maxtablesize*0x400-sizeof(TDataBlock)) / pih->px_recordsize;
 		blocknumber = 2; /* The first block contains the level 2 entries */
@@ -1245,7 +1249,7 @@ px_get_record_pos(pxdoc_t *pxdoc, int recno, int *deleted, pxdatablockinfo_t *px
 			datasize = blocksize; //get_short_le((char *) &datablock.addDataSize);
 		else
 			datasize = pxh->px_maxtablesize*0x400-sizeof(TDataBlock)-pxh->px_recordsize;
-		if(blocksize > pxh->px_maxtablesize*0x400-sizeof(TDataBlock)-pxh->px_recordsize) {
+		if(blocksize > pxh->px_maxtablesize*0x400-(int)sizeof(TDataBlock)-pxh->px_recordsize) {
 			/* setting blocksize to -1 means that this block contains no valid
 			 * records. All records are deleted. The -1 is later used to set
 			 * 'deleted' on the proper value.
@@ -1263,7 +1267,7 @@ px_get_record_pos(pxdoc_t *pxdoc, int recno, int *deleted, pxdatablockinfo_t *px
 		 * them, because the data is still there, but considered to be
 		 * deleted.
 		 */
-		if ((datasize+pxh->px_recordsize) > (pxh->px_maxtablesize*0x400-sizeof(TDataBlock))) {
+		if ((datasize+pxh->px_recordsize) > (pxh->px_maxtablesize*0x400-(int)sizeof(TDataBlock))) {
 //			printf("Size of data block %d as set in its header is to large: %d (%3.2f records)\n", get_short_le(&datablock.prevBlock), datasize, (float) datasize/pxh->px_recordsize + 1);
 			/* Set the number of the next block */
 			blocknumber = get_short_le((char *) &datablock.nextBlock);
@@ -1484,6 +1488,9 @@ PX_close(pxdoc_t *pxdoc) {
 		return;
 	}
 
+	/* Write modified cache block */
+	px_flush(pxdoc, pxdoc->px_stream);
+
 	if(pxdoc->px_blob) {
 		PX_close_blob(pxdoc->px_blob);
 		pxdoc->px_blob = NULL;
@@ -1571,6 +1578,10 @@ PX_delete(pxdoc_t *pxdoc) {
 		pxdoc->px_indexdatalen = 0;
 	}
 
+	/* Free the memory for the block cache */
+	if(pxdoc->curblock) {
+		pxdoc->free(pxdoc, pxdoc->curblock);
+	}
 	pxdoc->free(pxdoc, pxdoc);
 }
 /* }}} */
@@ -1825,27 +1836,23 @@ PX_open_blob_fp(pxblob_t *pxblob, FILE *fp) {
 		return -1;
 	}
 
-	if(NULL == (pxs = pxdoc->malloc(pxdoc, sizeof(pxstream_t), _("Allocate memory for io stream of blob file.")))) {
-		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for io stream of blob file."));
+	if(NULL == (pxs = px_stream_new_file(pxdoc, pxfFileRead, px_false, fp))) {
+		px_error(pxdoc, PX_MemoryError, _("Could not create new file io stream."));
 		return -1;
 	}
-	pxs->type = pxfIOFile;
-	pxs->mode = pxfFileRead;
-	pxs->close = px_false;
-	pxs->s.fp = fp;
-	
-	pxblob->read = px_fread;
-	pxblob->seek = px_fseek;
-	pxblob->tell = px_ftell;
-	pxblob->write = px_fwrite;
+
+	pxblob->mb_stream = pxs;
+
+	pxblob->read = px_mb_read;
+	pxblob->seek = px_mb_seek;
+	pxblob->tell = px_mb_tell;
+	pxblob->write = px_mb_write;
 
 	if((pxblob->mb_head = get_mb_head(pxblob, pxs)) == NULL) {
 		px_error(pxdoc, PX_RuntimeError, _("Unable to get header of blob file."));
 		return -1;
 	}
 
-	pxblob->mb_stream = pxs;
-
 	return(0);
 }
 /* }}} */
@@ -1890,19 +1897,17 @@ PX_create_blob_fp(pxblob_t *pxblob, FILE *fp) {
 		return -1;
 	}
 
-	if(NULL == (pxs = pxdoc->malloc(pxdoc, sizeof(pxstream_t), _("Allocate memory for io stream of blob file.")))) {
-		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for io stream of blob file."));
+	if(NULL == (pxs = px_stream_new_file(pxdoc, pxfFileWrite, px_false, fp))) {
+		px_error(pxdoc, PX_MemoryError, _("Could not create new file io stream."));
 		return -1;
 	}
-	pxs->type = pxfIOFile;
-	pxs->mode = pxfFileWrite;
-	pxs->close = px_false;
-	pxs->s.fp = fp;
-	
-	pxblob->read = px_fread;
-	pxblob->seek = px_fseek;
-	pxblob->tell = px_ftell;
-	pxblob->write = px_fwrite;
+
+	pxblob->mb_stream = pxs;
+
+	pxblob->read = px_mb_read;
+	pxblob->seek = px_mb_seek;
+	pxblob->tell = px_mb_tell;
+	pxblob->write = px_mb_write;
 
 	if(NULL == (mbh = pxdoc->malloc(pxdoc, sizeof(mbhead_t), _("Allocate memory for header of blob file.")))) {
 		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for header of blob file."));
@@ -1915,7 +1920,6 @@ PX_create_blob_fp(pxblob_t *pxblob, FILE *fp) {
 	}
 
 	pxblob->mb_head = mbh;
-	pxblob->mb_stream = pxs;
 	pxblob->used_datablocks = 0;
 
 	return(0);
@@ -2146,13 +2150,13 @@ _px_read_blobdata(pxblob_t *pxblob, const char *data, int len, int hsize, int *m
 		}
 //		fprintf(stderr, "offset=%ld ", offset);
 
-		if((ret = pxblob->seek(pxdoc, pxblob->mb_stream, offset, SEEK_SET)) < 0) {
+		if((ret = pxblob->seek(pxblob, pxblob->mb_stream, offset, SEEK_SET)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
 			return NULL;
 		}
 
 		/* Just read the first 3 Bytes because they are common for all block */
-		if((ret = pxblob->read(pxdoc, pxblob->mb_stream, 3, head)) < 0) {
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, 3, head)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not read head of blob data."));
 			return NULL;
 		}
@@ -2171,12 +2175,12 @@ _px_read_blobdata(pxblob_t *pxblob, const char *data, int len, int hsize, int *m
 				return NULL;
 			}
 			/* Read the remaining 6/14 bytes from the header */
-			if((ret = pxblob->read(pxdoc, pxblob->mb_stream, hsize-3, head)) < 0) {
+			if((ret = pxblob->read(pxblob, pxblob->mb_stream, hsize-3, head)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not read remaining head of single data block."));
 				return NULL;
 			}
-			if(size != get_long_le(&head[0])) {
-				px_error(pxdoc, PX_RuntimeError, _("Blob does not have expected size (%d != %d)."), size, get_long_le(&head[0]));
+			if(size != get_long_le((const char *) &head[0])) {
+				px_error(pxdoc, PX_RuntimeError, _("Blob does not have expected size (%d != %d)."), size, get_long_le((const char *)&head[0]));
 				return(NULL);
 			}
 			/* We may check for identical modificatio number as well, if it
@@ -2188,23 +2192,23 @@ _px_read_blobdata(pxblob_t *pxblob, const char *data, int len, int hsize, int *m
 				return(NULL);
 			}
 
-			if((ret = pxdoc->read(pxdoc, pxblob->mb_stream, *blobsize, blobdata)) < 0) {
+			if((ret = pxblob->read(pxblob, pxblob->mb_stream, *blobsize, blobdata)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not read all blob data."));
 				return NULL;
 			}
 		} else if(head[0] == 3) { /* Reading data from a block type 3 */
 			/* Read the remaining 9 bytes from the header */
-			if((ret = pxblob->read(pxdoc, pxblob->mb_stream, 9, head)) < 0) {
+			if((ret = pxblob->read(pxblob, pxblob->mb_stream, 9, head)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not read remaining head of suballocated block."));
 				return NULL;
 			}
 			/* Goto the blob pointer with the passed index */
-			if((ret = pxblob->seek(pxdoc, pxblob->mb_stream, offset+12+index*5, SEEK_SET)) < 0) {
+			if((ret = pxblob->seek(pxblob, pxblob->mb_stream, offset+12+index*5, SEEK_SET)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not fseek blob pointer."));
 				return NULL;
 			}
 			/* Read the blob pointer */
-			if((ret = pxblob->read(pxdoc, pxblob->mb_stream, 5, head)) < 0) {
+			if((ret = pxblob->read(pxblob, pxblob->mb_stream, 5, head)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not read blob pointer."));
 				return NULL;
 			}
@@ -2217,11 +2221,11 @@ _px_read_blobdata(pxblob_t *pxblob, const char *data, int len, int hsize, int *m
 				return(NULL);
 			}
 			/* Goto the start of the blob */
-			if((ret = pxblob->seek(pxdoc, pxblob->mb_stream, offset+head[0]*16, SEEK_SET)) < 0) {
+			if((ret = pxblob->seek(pxblob, pxblob->mb_stream, offset+head[0]*16, SEEK_SET)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
 				return NULL;
 			}
-			if((ret = pxblob->read(pxdoc, pxblob->mb_stream, size, blobdata)) < 0) {
+			if((ret = pxblob->read(pxblob, pxblob->mb_stream, size, blobdata)) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not read all blob data."));
 				return NULL;
 			}
@@ -2284,7 +2288,7 @@ PX_get_data_alpha(pxdoc_t *pxdoc, char *data, int len, char **value) {
 		optr = obuf = (char *) malloc(olen);
 		iptr = data;
 		ilen = 0;
-		while(iptr[ilen] != '\0' && ilen < len)
+		while(iptr[ilen] != '\0' && ilen < (size_t) len)
 			ilen++;
 //		printf("data(%d) = '%s'\n", ilen, data);
 //		printf("obuf(%d) = '%s'\n", olen, obuf);
@@ -2506,7 +2510,7 @@ static int
 _px_get_data_blob(pxdoc_t *pxdoc, const char *data, int len, int hsize, int *mod, int *blobsize, char **value) {
 	int ret;
 	char *blobdata = NULL;
-	char head[20];
+	unsigned char head[20];
 	pxblob_t *pxblob = pxdoc->px_blob;
 	size_t size, offset, mod_nr, index;
 	int leader = len - 10;
@@ -2561,14 +2565,14 @@ _px_get_data_blob(pxdoc_t *pxdoc, const char *data, int len, int hsize, int *mod
 	}
 //		fprintf(stderr, "offset=%ld ", offset);
 
-	if((ret = pxdoc->seek(pxdoc, pxblob->mb_stream, offset, SEEK_SET)) < 0) {
+	if((ret = pxblob->seek(pxblob, pxblob->mb_stream, offset, SEEK_SET)) < 0) {
 		px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
 		*value = NULL;
 		return -1;
 	}
 
 	/* Just read the first 3 Bytes because they are common for all block */
-	if((ret = pxblob->read(pxdoc, pxblob->mb_stream, 3, head)) < 0) {
+	if((ret = pxblob->read(pxblob, pxblob->mb_stream, 3, head)) < 0) {
 		px_error(pxdoc, PX_RuntimeError, _("Could not read head of blob data."));
 		*value = NULL;
 		return -1;
@@ -2591,7 +2595,7 @@ _px_get_data_blob(pxdoc_t *pxdoc, const char *data, int len, int hsize, int *mod
 			return -1;
 		}
 		/* Read the remaining 6 bytes from the header */
-		if((ret = pxblob->read(pxdoc, pxblob->mb_stream, hsize-3, head)) < 0) {
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, hsize-3, head)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not read remaining head of single data block."));
 			*value = NULL;
 			return -1;
@@ -2611,26 +2615,26 @@ _px_get_data_blob(pxdoc_t *pxdoc, const char *data, int len, int hsize, int *mod
 			return -1;
 		}
 
-		if((ret = pxblob->read(pxdoc, pxblob->mb_stream, *blobsize, blobdata)) < 0) {
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, *blobsize, blobdata)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not read all blob data."));
 			*value = NULL;
 			return -1;
 		}
 	} else if(head[0] == 3) { /* Reading data from a block type 3 */
 		/* Read the remaining 9 bytes from the header */
-		if((ret = pxblob->read(pxdoc, pxblob->mb_stream, 9, head)) < 0) {
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, 9, head)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not read remaining head of suballocated block."));
 			*value = NULL;
 			return -1;
 		}
 		/* Goto the blob pointer with the passed index */
-		if((ret = pxblob->seek(pxdoc, pxblob->mb_stream, offset+12+index*5, SEEK_SET)) < 0) {
+		if((ret = pxblob->seek(pxblob, pxblob->mb_stream, offset+12+index*5, SEEK_SET)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not fseek blob pointer."));
 			*value = NULL;
 			return -1;
 		}
 		/* Read the blob pointer */
-		if((ret = pxblob->read(pxdoc, pxblob->mb_stream, 5, head)) < 0) {
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, 5, head)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not read blob pointer."));
 			*value = NULL;
 			return -1;
@@ -2646,12 +2650,12 @@ _px_get_data_blob(pxdoc_t *pxdoc, const char *data, int len, int hsize, int *mod
 			return -1;
 		}
 		/* Goto the start of the blob */
-		if((ret = pxblob->seek(pxdoc, pxblob->mb_stream, offset+head[0]*16, SEEK_SET)) < 0) {
+		if((ret = pxblob->seek(pxblob, pxblob->mb_stream, offset+head[0]*16, SEEK_SET)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not fseek start of blob."));
 			*value = NULL;
 			return -1;
 		}
-		if((ret = pxblob->read(pxdoc, pxblob->mb_stream, size, blobdata)) < 0) {
+		if((ret = pxblob->read(pxblob, pxblob->mb_stream, size, blobdata)) < 0) {
 			px_error(pxdoc, PX_RuntimeError, _("Could not read all blob data."));
 			*value = NULL;
 			return -1;
@@ -2727,7 +2731,7 @@ PX_put_data_alpha(pxdoc_t *pxdoc, char *data, int len, char *value) {
 		obuf = value;
 	}
 
-	memcpy(data, obuf, olen < len ? olen : len);
+	memcpy(data, obuf, olen < (size_t) len ? olen : len);
 
 	if(pxdoc->targetencoding != NULL) {
 		free(obuf);
@@ -2916,7 +2920,7 @@ _px_put_data_blob(pxdoc_t *pxdoc, const char *data, int len, char *value, int va
 			int used_blocks;
 
 //			fprintf(stderr, "Blob goes into type 2 block\n");
-			if(pxblob->seek(pxdoc, pxs, (pxblob->used_datablocks+1)*4096, SEEK_SET) < 0) {
+			if(pxblob->seek(pxblob, pxs, (pxblob->used_datablocks+1)*4096, SEEK_SET) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not go to the begining of the first free block in the blob file."));
 				return -1;
 			}
@@ -2933,12 +2937,12 @@ _px_put_data_blob(pxdoc_t *pxdoc, const char *data, int len, char *value, int va
 			put_short_le((char *) &mbbh.modNr, ++pxblob->mb_head->modcount);
 
 			/* Write the header of the blob */
-			if(pxblob->write(pxdoc, pxs, sizeof(TMbBlockHeader2), &mbbh) < 1) {
+			if(pxblob->write(pxblob, pxs, sizeof(TMbBlockHeader2), &mbbh) < 1) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not write header of blob data to file."));
 				return -1;
 			}
 			/* Write the blob itself */
-			if(pxblob->write(pxdoc, pxs, valuelen, value) < 1) {
+			if(pxblob->write(pxblob, pxs, valuelen, value) < 1) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not write blob data to file."));
 				return -1;
 			}
@@ -2947,13 +2951,13 @@ _px_put_data_blob(pxdoc_t *pxdoc, const char *data, int len, char *value, int va
 			pxblob->used_datablocks += used_blocks;
 		} else { /* Block of type 3 */
 			TMbBlockHeader3Table mbbhtab;
-			fprintf(stderr, "Blob goes into type 3 block\n");
+//			fprintf(stderr, "Blob goes into type 3 block\n");
 			/* Do we have subblock already? Does the block has enough space? */
 			if(pxblob->subblockoffset == 0 || ((pxblob->subblockfree*16) < valuelen)) {
 				TMbBlockHeader3 mbbh;
 				int i, nullint=0;
 
-				if(pxblob->seek(pxdoc, pxs, (pxblob->used_datablocks+1)*4096, SEEK_SET) < 0) {
+				if(pxblob->seek(pxblob, pxs, (pxblob->used_datablocks+1)*4096, SEEK_SET) < 0) {
 					px_error(pxdoc, PX_RuntimeError, _("Could not go to the begining of the first free block in the blob file."));
 					return -1;
 				}
@@ -2962,12 +2966,12 @@ _px_put_data_blob(pxdoc_t *pxdoc, const char *data, int len, char *value, int va
 				mbbh.type = 3;
 				put_short_le((char *) &mbbh.numBlocks, 1);
 				/* Write the header of the blob */
-				if(pxblob->write(pxdoc, pxs, sizeof(TMbBlockHeader3), &mbbh) < 1) {
+				if(pxblob->write(pxblob, pxs, sizeof(TMbBlockHeader3), &mbbh) < 1) {
 					px_error(pxdoc, PX_RuntimeError, _("Could not write header of blob data to file."));
 					return -1;
 				}
 				for(i=0; i<4096-sizeof(TMbBlockHeader3); i++) {
-					if(pxblob->write(pxdoc, pxs, 1, &nullint) < 1) {
+					if(pxblob->write(pxblob, pxs, 1, &nullint) < 1) {
 						px_error(pxdoc, PX_RuntimeError, _("Could not write remaining of a type 3 block."));
 						return -1;
 					}
@@ -2983,7 +2987,7 @@ _px_put_data_blob(pxdoc_t *pxdoc, const char *data, int len, char *value, int va
 			 * First write the table entry pointing to the blob data. The table is
 			 * filled from the end to the beginning.
 			 */
-			if(pxblob->seek(pxdoc, pxs, (pxblob->subblockoffset)*4096+sizeof(TMbBlockHeader3)+(63-pxblob->subblockblobcount)*5, SEEK_SET) < 0) {
+			if(pxblob->seek(pxblob, pxs, (pxblob->subblockoffset)*4096+sizeof(TMbBlockHeader3)+(63-pxblob->subblockblobcount)*5, SEEK_SET) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not go to table entry for the blob data."));
 				return -1;
 			}
@@ -3000,16 +3004,16 @@ _px_put_data_blob(pxdoc_t *pxdoc, const char *data, int len, char *value, int va
 			put_short_le((char *) &mbbhtab.modNr, pxblob->mb_head->modcount+1);
 			mbbhtab.lengthmod = (valuelen % 16) == 0 ? 16 : (valuelen % 16);
 			/* Write the blob table entry */
-			if(pxblob->write(pxdoc, pxs, sizeof(TMbBlockHeader3Table), &mbbhtab) < 1) {
+			if(pxblob->write(pxblob, pxs, sizeof(TMbBlockHeader3Table), &mbbhtab) < 1) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not write table entry for blob data to file."));
 				return -1;
 			}
 			/* Write the blob itself */
-			if(pxblob->seek(pxdoc, pxs, pxblob->subblockoffset*4096+mbbhtab.offset*16, SEEK_SET) < 0) {
+			if(pxblob->seek(pxblob, pxs, pxblob->subblockoffset*4096+mbbhtab.offset*16, SEEK_SET) < 0) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not go to the begining of the slot for the blob."));
 				return -1;
 			}
-			if(pxblob->write(pxdoc, pxs, valuelen, value) < 1) {
+			if(pxblob->write(pxblob, pxs, valuelen, value) < 1) {
 				px_error(pxdoc, PX_RuntimeError, _("Could not write blob data to file."));
 				return -1;
 			}
@@ -3101,7 +3105,7 @@ PX_timestamp2string(pxdoc_t *pxdoc, double value, const char *format) {
 
 		}
 	}
-	if(NULL == (str = pxdoc->malloc(pxdoc, size, _("Allocate memory for timestamp string.")))) {
+	if(NULL == (str = pxdoc->malloc(pxdoc, size+1, _("Allocate memory for timestamp string.")))) {
 		px_error(pxdoc, PX_MemoryError, _("Could not allocate memory for timestamp string."));
 		return NULL;
 	}
@@ -3116,7 +3120,7 @@ PX_timestamp2string(pxdoc_t *pxdoc, double value, const char *format) {
 					strcat(str, ch);
 				}
 			case 'Y':		/* year, numeric, 4 digits */
-				sprintf(tmp_buff, "%d", ta.tm_year);  /* SAFE */
+				sprintf(tmp_buff, "%04d", ta.tm_year);  /* SAFE */
 				strcat(str, tmp_buff);
 				break;
 			case 'z':		/* day (of the year) */
